#! /usr/bin/python
## -*- coding: utf-8 -*-

##
##  HARK Middleware
##  Copyright (C) 2013-2014 by Honda Research Institute Japan Co., Ltd.
##
##  $Id:$
##

import threading

from harkmw.HarkError import HARKError
from harkmw.logic.link import Link
from harkmw.logic.mqtt_node import MQTTNode

import harkmwnative

import os
debug = (bool(os.environ['HARK_DEBUG'])) if 'HARK_DEBUG' in os.environ else False

class ProxyTerminal :
    def __init__(self, node, terminal) :
        self.node     = node
        self.terminal = terminal
        self.proxy    = None

    def __str__(self) :
        return "<ProxyTerminal: node=%s/%s, terminal=%s, proxy=%s>" % (self.node.type, self.node.name, self.terminal, self.proxy)

class Proxy :
    """ Base-class of Network, NativeProxy, Proxy
    """
    def __init__(self, process) :
        self.native = harkmwnative.Proxy(self)
        self.process = process
        self.input_node  = None
        self.output_node = None
        self.count       = -1
        self.forced_count = -1

        self.buffer        = []
        self.buffer_offset = 0
        self.buffer_lock   = threading.Lock()

    def __str__(self) :
        return "<Proxy: input=%s, output=%s>" % (self.input_node, self.output_node)

    def connect_native_input(self, node, terminal) :
        """ この Proxy オブジェクトの入口側に、C++ 実装ノードの出口をつなぐ
        """
#       print "Proxy[%s].connect_native_input(node=%s/%s, terminal=%s)" % (self, node.type, node.name, terminal) 
        if self.input_node is not None :
            raise HARKError("Proxy node already connected")

        self.native.connect_input(node.native_node, terminal)
        self.input_node          = node
        self.input_terminal_name = terminal

    def connect_network_input(self, network, terminal) :
        """ この Proxy オブジェクトの入口側に、ネットワークの出口をつなぐ
        """
#       print "Proxy[%s].connect_network_input(network=%s, terminal=%s)" % (self, network.name, terminal) 
        if self.input_node is not None :
            raise HARKError("Proxy node already connected")

        proxyTerminal = network.output_terminals[terminal]
#       print "Node: %s" % proxyTerminal.node
        self.native.connect_input(proxyTerminal.node.native_node, proxyTerminal.terminal)
        self.input_node           = proxyTerminal.node
        self.input_terminal_name  = proxyTerminal.terminal

    def connect_node_input(self, node, terminal) :
        """ この Proxy オブジェクトの入口側に、C++ 実装でないノードの出口をつなぐ
        """
#       print "Proxy[%s].connect_native_input(node=%s/%s, terminal=%s)" % (self, node.type, node.name, terminal) 
        if self.input_node is not None :
            raise HARKError("Proxy node already connected")

        self.input_node          = node
        self.input_terminal_name = terminal

    def connect_native_output(self, node, terminal) :
        """ この Proxy オブジェクトの出口側に、C++ 実装ノードの入口をつなぐ
        """
#       print "Proxy[%s].connect_native_output(node=%s/%s, terminal=%s)" % (self, node.type, node.name, terminal) 
        if self.output_node is not None :
            raise HARKError("Proxy node already connected")

        self.native.connect_output(node.native_node, terminal)
        self.output_node          = node
        self.output_terminal_name = terminal

    def connect_network_output(self, network, terminal) :
        """ この Proxy オブジェクトの出口側に、ネットワークの入口をつなぐ
        """
#       print "Proxy[%s].connect_network_output(network=%s, terminal=%s)" % (self, network.name, terminal) 
        if self.output_node is not None :
            raise HARKError("Proxy node already connected")

        proxyTerminal = network.input_terminals[terminal]
#       print "Node: %s" % proxyTerminal.node
        self.native.connect_output(proxyTerminal.node.native_node, proxyTerminal.terminal)
        self.output_node           = proxyTerminal.node
        self.output_terminal_name  = proxyTerminal.terminal

    def connect_node_output(self, node, terminal) :
        """ この Proxy オブジェクトの出口側に、C++ 実装でないノードの入口をつなぐ
        """
#       print "Proxy[%s].connect_native_output(node=%s/%s, terminal=%s)" % (self, node.type, node.name, terminal) 
        if self.output_node is not None :
            raise HARKError("Proxy node already connected")

        self.output_node          = node
        self.output_terminal_name = terminal

    def request(self, process):
        look_windows = [(0, 0)]
        if self.output_node is not None:
            for (output_name, output) in self.output_node.outputs.items() :
                if debug:
                    print("output_name: {}, output={}, node={}".format(output_name, output, self.output_node))
                look_windows.append(self.output_node.get_output_look_window(output_name))
        look_window = max(look_windows)
        look_window = [sum(x) for x in zip(look_window, self.output_node.get_input_look_window(self.output_terminal_name))]

        if debug:
            print("Proxy: request: self={}, look_window={}, look_windows={}".format(self, look_window, look_windows))
        self.set_look_window(look_window)
        if self.input_node is not None:
            self.input_node.request(process)

    def initialize(self, process) :
        #print("Initialize: proxy: {}".format(self))
        self.native.initialize()

#        print "---- %s" % ("%s/%s" % (self.input_node.get_topic(), self.input_terminal_name))
#        print self.output_node
#        print self.input_node.network
#        print self.input_node.network.nodes.get('NETWORK_CONDITION')

#        if isinstance(self.input_node, MQTTNode) and ((not isinstance(self.output_node, MQTTNode) and not self.output_node.type == "Collector") \
#            or (self.input_node.network is not None and self.output_node == self.input_node.network.nodes.get('NETWORK_CONDITION'))) :
        if isinstance(self.input_node, MQTTNode) and ((not isinstance(self.output_node, MQTTNode) and not self.output_node.type == "Collector")) :
            """ Proxy の入口のノードはリモート実行、あるいは CONDITION 出力 """
            topic = "%s/%s" % (self.input_node.get_topic(), self.input_terminal_name)

#           print "Subscribe: %s -> %s/%s" % (topic, self.output_node.name, self.output_terminal_name)
            self.process.subscribe(topic, self)
            self.input_node.network.mqtt_proxies.append(self)

    def push_data(self, count, process, body=None) :
#        print "Proxy<%s>: push_data, count=%d/%d" % (self, self.count, count)
#       print "Source: %s" % self.input_node
#       print "Sink: %s" % self.output_node

#        if isinstance(self.output_node, MQTTNode) \
#            or (self.input_node.network is not None and self.output_node == self.input_node.network.nodes.get('NETWORK_CONDITION')) :
        if isinstance(self.output_node, MQTTNode) :
            """ Proxy の出口のノードはリモート実行 """
            data = self.native.get_data(count)
            topic = "%s/%s" % (self.input_node.get_topic(), self.input_terminal_name)
            
#           print "PublishData: %s, %d, %s" % (topic, count, data)
            self.process.publish(topic, count, data)
            
        else :
            if body is not None:
#               print "Body: %s" % body
#               print "Body: %s" % dir(body)
#               print "Body: %s" % body.__class__
                self.native.set_data(count, body)
            count = self.native.push_data(count)
#            print "native.push_data: ok"
            self.count = count

            if self.output_node.type in ("Collector", "HARKMW_Collector") and self.output_node.name[-6:] == "_input" :
                self.output_node.network.push_data(input, count, process)
            else :
                count = self.output_node.push_data(self, count, process)
        return count

    def get_data(self, count) :
#       print "Proxy: get_output: %d" % count
        if isinstance(self.input_node, NativeNode) :
            return native.get_data(count)
        else :
            return self.input_node.get_output(self.input_terminal_name, count)

    def get_input_node(self) :
        return self.input_node

    def get_output_node(self) :
        return self.output_node

    def get_input_terminal_name(self) :
        return self.input_terminal_name

    def get_output_terminal_name(self) :
        return self.output_terminal_name

    def force_count(self, count) :
        self.forced_count = count
        self.native.force_count(count)

#    def is_count_available(self, count) :
#        lookAhead = self.get_look_window()[0]
#
##       print "Proxy: %s(%d/%d), Count: %d, lookAhead: %d" % (self, self.forced_count, self.count, count, lookAhead)
#        if self.forced_count >= 0 :
#            return self.count >= self.forced_count + lookAhead
#        else :
#            return self.count >= count + lookAhead

    def get_look_window(self) :
        return self.native.get_look_window()

    def set_look_window(self, look_window) :
        return self.native.set_look_window(look_window[0], look_window[1])

    def get_delay(self) :
        if self.input_node is None :
            """ 他のプロセスでノードが動いている場合
            """
            return 0

        elif self.input_node.type == "Collector" and self.input_node.name[-6:] == "_input" :
            return self.native.get_look_window()[0]

        else :
            #print "self.input_node.get_delay() + self.native.get_look_ahead() = %d + %d" % (self.input_node.get_delay(), self.native.get_look_ahead())
            return self.input_node.get_delay() + self.native.get_look_window()[0]

# end of file
