/* -*- mode: C++; coding: utf-8 -*- */
/*
 *
 *  HARK Middleware
 *  Copyright (C) 2013-2014 by Honda Research Institute Japan Co., Ltd.
 *
 *  $Id:$
 */

#include "libharkmw.h"
#include "proxy.h"

#include <hark-base/BufferedNode.h>
#include <hark-base/Collector.h>
#include <hark-base/object_param.h>
#include <hark-base/Vector.h>

#include "node.h"
#include "type.h"

namespace HARKMW
{
    class ProxyNode;

    struct Proxy
    {
        PyObject_HEAD

        PyObject  *proxyObject;
        ProxyNode *proxyNode;
    };

    class ProxyNode : public HARK::BufferedNode
    {
    private:
        int count;
        Proxy *proxy;

    public:
        int input_id;
        int output_id;

    public:
        ProxyNode(std::string nodeName, const HARK::ParameterSet &params, Proxy *proxy) :
            HARK::BufferedNode(nodeName, params)
        {
            this->count     = -1;
            this->input_id  = this->addInput ("I/O");
            this->output_id = this->addOutput("I/O");
            this->proxy = proxy;
        }

    public:
        void force_count(int count)
        {
            this->count = count;
        }

        virtual void calculate(int output_id, int count, HARK::Buffer &buf)
        {
            //int target_count = this->count >= 0 ? this->count : count;
            int target_count = count;

            HARK::NodeInput &input = inputs[this->input_id];
            HARK::ObjectRef inputValue;
            if(input.node){
                inputValue = input.node->getOutput(input.outputID, target_count);
            }else{
                PyObject *input_node = PyObject_GetAttrString(this->proxy->proxyObject, "input_node");
                PyObject *input_terminal_name = PyObject_GetAttrString(this->proxy->proxyObject, "input_terminal_name");

                char method_name[] = "get_output";
                char arg_format[]  = "oi";
                inputValue = fromPythonObject(PyObject_CallMethod(input_node, method_name, arg_format, input_terminal_name, target_count));
            }
            buf[count] = inputValue;
        }
        
        virtual HARK::ObjectRef getOutput(int output_id, int count)
        {
            return HARK::BufferedNode::getOutput(output_id, count);
        }

        void setOutput(int output_id, int count, HARK::ObjectRef &data)
        {
            HARK::Buffer &outBuffer = *(outputs[output_id].buffer);
            outBuffer[count] = data;
        }

        virtual void initialize()
        {
            this->HARK::BufferedNode::initialize();
        }

    };

    static PyObject *
    Proxy_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
    {
        PyObject *_self = (PyObject *)type->tp_alloc(type, 0);
        
        if(_self != NULL){
            Proxy *self = reinterpret_cast<Proxy *>(_self);

            HARK::ParameterSet emptyParams;
            self->proxyObject = NULL;
            self->proxyNode   = new ProxyNode(std::string("ProxyNode"), emptyParams, self);
        }

        return _self;
    }

    static void
    Proxy_dealloc(Proxy* self)
    {
        Py_XDECREF(self->proxyObject);
        if(self->proxyNode){
            delete self->proxyNode;
        }
        Py_TYPE(self)->tp_free((PyObject*)self);
    }

    static int
    Proxy_init(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Proxy *self = reinterpret_cast<Proxy *>(_self);
        PyObject *proxyObj = NULL;

        /* Parse args */
        static char *kwlist[] = {"proxy", NULL};
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "O", kwlist, &proxyObj)){
            return -1;
        }

        Py_INCREF(proxyObj);
        Py_XDECREF(self->proxyObject);

        self->proxyObject = proxyObj;
        return 0;
    }

    static PyObject *
    Proxy_connect_input(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Proxy *self = reinterpret_cast<Proxy *>(_self);

        /* Parse args */
        PyObject *nativeNodeObj = NULL, *terminalNameObj = NULL;
        static char *kwlist[] = {"node", "terminal_name", NULL};
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "OO", kwlist, &nativeNodeObj, &terminalNameObj)){
            return NULL;
        }

        if(!PyObject_TypeCheck(nativeNodeObj, &NodeType)){
            Py_RETURN_NONE;
        }

        Node *_node = reinterpret_cast<Node *>(nativeNodeObj);
        HARK::Node *node = _node->node;

        try{
            self->proxyNode->connectToNode(std::string("I/O"), node, std::string(PyStr_AsUTF8(terminalNameObj)));
	}HARKMW_CATCH_EXCEPTION();

        Py_RETURN_NONE;
    }

    static PyObject *
    Proxy_connect_output(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Proxy *self = reinterpret_cast<Proxy *>(_self);

        /* Parse args */
        PyObject *nativeNodeObj = NULL, *terminalNameObj = NULL;
        static char *kwlist[] = {"node", "terminal_name", NULL};
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "OO", kwlist, &nativeNodeObj, &terminalNameObj)){
            return NULL;
        }

        if(!PyObject_TypeCheck(nativeNodeObj, &NodeType)){
            Py_RETURN_NONE;
        }

        Node *_node = reinterpret_cast<Node *>(nativeNodeObj);
        HARK::Node *node = _node->node;

        try{
            HARK::BufferedNode *bufferedNode = dynamic_cast<HARK::BufferedNode *>(node);
            if(bufferedNode){
                std::string name = std::string(PyStr_AsUTF8(terminalNameObj));
                int inputId = -1;
                for(std::vector<HARK::NodeInput>::const_iterator it = node->inputs.begin(); it != node->inputs.end(); ++it){
                    if((*it).name == name){
                        inputId = (it - node->inputs.begin());
                        break;
                    }
                }
                if(inputId >= 0){
                    self->proxyNode->outputs[self->proxyNode->output_id].lookAhead = bufferedNode->inputsCache[inputId].lookAhead;
                    self->proxyNode->outputs[self->proxyNode->output_id].lookBack = bufferedNode->inputsCache[inputId].lookBack;
                }
            }

            node->connectToNode(std::string(PyStr_AsUTF8(terminalNameObj)), self->proxyNode, std::string("I/O"));
	}HARKMW_CATCH_EXCEPTION();

        Py_RETURN_NONE;
    }

    static PyObject *
    Proxy_initialize(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Proxy *self = reinterpret_cast<Proxy *>(_self);

        /* Parse args */
        self->proxyNode->initialize();

        Py_RETURN_NONE;
    }

    static PyObject *
    Proxy_push_data(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Proxy *self = reinterpret_cast<Proxy *>(_self);

        /* Parse args */
        int count = 0;
        static char *kwlist[] = {"count", NULL};
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "i", kwlist, &count)){
            return NULL;
        }

        try{
            self->proxyNode->getOutput(self->proxyNode->output_id, count);
	}HARKMW_CATCH_EXCEPTION();

        return Py_BuildValue("i", count);
        //Py_RETURN_TRUE;
    }

    static PyObject *
    Proxy_get_data(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Proxy *self = reinterpret_cast<Proxy *>(_self);

        /* Parse args */
        int count = 0;
        static char *kwlist[] = {"count", NULL};
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "i", kwlist, &count)){
            return NULL;
        }

        try{
            HARK::ObjectRef out = self->proxyNode->getOutput(0, count);

            return toPythonObject(out);
            /*
            if(HARK::dereference_cast<bool>(out) == false){
                fprintf(stdout, "Returning false\n");
                Py_RETURN_FALSE;
            }else{
                fprintf(stdout, "Returning true\n");
                Py_RETURN_TRUE;
            }
            */

        }catch(HARK::BaseException *ex){
            ex->print();
            delete ex;
        }

        Py_RETURN_NONE;
    }

    static PyObject *
    Proxy_set_data(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Proxy *self = reinterpret_cast<Proxy *>(_self);

        /* Parse args */
        int count = 0;
        PyObject *dataObject = NULL;
        static char *kwlist[] = {"count", "data", NULL};
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "iO", kwlist, &count, &dataObject)){
            return NULL;
        }

        try{
            HARK::ObjectRef value = fromPythonObject(dataObject);
            self->proxyNode->setOutput(0, count, value);

	}HARKMW_CATCH_EXCEPTION();

        Py_RETURN_NONE;
    }

    static PyObject *
    Proxy_get_look_window(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Proxy *self = reinterpret_cast<Proxy *>(_self);

        /* Parse args */
        static char *kwlist[] = {NULL};
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "", kwlist)){
            return NULL;
        }

	//std::cout << "get: " << self->proxyNode->outputs[self->proxyNode->output_id].lookAhead << ", " << self->proxyNode->outputs[self->proxyNode->output_id].lookBack << std::endl;
        return Py_BuildValue("(ii)",
			     self->proxyNode->outputs[self->proxyNode->output_id].lookAhead,
			     self->proxyNode->outputs[self->proxyNode->output_id].lookBack);
    }

    static PyObject *
    Proxy_set_look_window(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Proxy *self = reinterpret_cast<Proxy *>(_self);

        /* Parse args */
        int look_ahead = 0, look_back = 0;
        static char *kwlist[] = {"lookAhead", "lookBack", NULL};
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "ii", kwlist, &look_ahead, &look_back)){
            return NULL;
        }

        self->proxyNode->outputs[self->proxyNode->output_id].lookAhead = look_ahead;
        self->proxyNode->outputs[self->proxyNode->output_id].lookBack  = look_back;
	//std::cout << "set: " << look_ahead << ", " << look_back << std::endl;

        Py_RETURN_NONE;
    }

    static PyObject *
    Proxy_force_count(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Proxy *self = reinterpret_cast<Proxy *>(_self);

        /* Parse args */
        int count = 0;
        static char *kwlist[] = {"count", NULL};
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "i", kwlist, &count)){
            return NULL;
        }

        try{
            self->proxyNode->force_count(count);

	}HARKMW_CATCH_EXCEPTION();

        Py_RETURN_NONE;
    }

    static PyMemberDef Proxy_members[] = {
        {NULL}  /* Sentinel */
    };

    static PyMethodDef Proxy_methods[] = {
        {"connect_input", (PyCFunction)Proxy_connect_input, METH_KEYWORDS | METH_VARARGS, ""},
        {"connect_output", (PyCFunction)Proxy_connect_output, METH_KEYWORDS | METH_VARARGS, ""},
        {"initialize", (PyCFunction)Proxy_initialize, METH_KEYWORDS | METH_VARARGS, ""},
        {"push_data", (PyCFunction)Proxy_push_data, METH_KEYWORDS | METH_VARARGS, ""},
        {"get_data", (PyCFunction)Proxy_get_data, METH_KEYWORDS | METH_VARARGS, ""},
        {"set_data", (PyCFunction)Proxy_set_data, METH_KEYWORDS | METH_VARARGS, ""},
        {"get_look_window", (PyCFunction)Proxy_get_look_window, METH_KEYWORDS | METH_VARARGS, ""},
        {"set_look_window", (PyCFunction)Proxy_set_look_window, METH_KEYWORDS | METH_VARARGS, ""},
        {"force_count", (PyCFunction)Proxy_force_count, METH_KEYWORDS | METH_VARARGS, ""},
        {NULL}  /* Sentinel */
    };

    PyTypeObject ProxyType = {
        PyVarObject_HEAD_INIT(NULL, 0)
        "harkmwnative.Proxy",             /*tp_name*/
        sizeof(HARKMW::Proxy),       /*tp_basicsize*/
        0,                         /*tp_itemsize*/
        (destructor)Proxy_dealloc, /*tp_dealloc*/
        0,                         /*tp_print*/
        0,                         /*tp_getattr*/
        0,                         /*tp_setattr*/
        0,                         /*tp_compare*/
        0,                         /*tp_repr*/
        0,                         /*tp_as_number*/
        0,                         /*tp_as_sequence*/
        0,                         /*tp_as_mapping*/
        0,                         /*tp_hash */
        0,                         /*tp_call*/
        0,                         /*tp_str*/
        0,                         /*tp_getattro*/
        0,                         /*tp_setattro*/
        0,                         /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
        "Proxy objects",            /* tp_doc */
        0,                               /* tp_traverse */
        0,                               /* tp_clear */
        0,                               /* tp_richcompare */
        0,                               /* tp_weaklistoffset */
        0,                               /* tp_iter */
        0,                               /* tp_iternext */
        Proxy_methods,              /* tp_methods */
        Proxy_members,              /* tp_members */
        0,                         /* tp_getset */
        0,                         /* tp_base */
        0,                         /* tp_dict */
        0,                         /* tp_descr_get */
        0,                         /* tp_descr_set */
        0,                         /* tp_dictoffset */
        (initproc)Proxy_init,       /* tp_init */
        0,                         /* tp_alloc */
        Proxy_new,                  /* tp_new */
    };
}

/* end of file */
